#!/bin/sh
#
# Shell script to build Asterix without using make.
#
# It should work with bash and busybox sh.
#
# Copyright (C) 2019  Pim Goossens
# See the README and COPYING files for details and license information.

CMETA=../cmeta
MCSS=../mcss
PROX=../prox

usage() {
    echo "Usage: $0 [-h] [-q | -qq] [target]"
    echo
    echo "  -h      Show this help text"
    echo "  -q      Don't show what's being checked for"
    echo "  -qq     Don't show what's being built or cleaned either"
    echo
    echo 'List of possible targets:'
    echo
    echo '  (*) asterix        Build asterix (and prox if necessary)'
    echo '      prox           Build prox (and mcss if necessary)'
    echo '      mcss           Build mcss (and cmeta if necessary)'
    echo '      cmeta          Build cmeta'
    echo '      prereq         Build just the prerequisites (cmeta, mcss, prox)'
    echo '      clean          Remove (clean) all generated Asterix files'
    echo '      clean-prereq   Clean all prerequisites'
    echo '      clean-all      Same as "clean" followed by "clean-prereq"'
    echo
    echo '(*) asterix is the default target if none is specified; it usually'
    echo '    suffices as it also builds all the other targets required'
    echo
    echo 'Note: if the build fails midway through, it may be necessary to run'
    echo 'the "clean-all" target before rebuilding. For more information,'
    echo 'see the README and bootstrap.txt files.'
    exit 0
}

die() {
    echo "$0: Error: $*" >&2
    exit 1
}

info() {
    [ "$loglevel" -ge 1 ] && echo "$*" >&2
}

info2() {
    [ "$loglevel" -ge 2 ] && echo "$*" >&2
}

info_build() {
    info "Building: $*"
}

info_clean() {
    info "Cleaning: $*"
}

info_prereq() {
    info2 "Checking prerequisites for: $*"
}

require() {
    info2 "Checking for required file: $1"
    [ -e "$1" ] || die "Missing file: $1"
}

# -------------------- cmeta -------------------

hex() {
    info_build 'hex'
    cd ../cmeta
    ./hex.sh < hex.hex > hex
    chmod +x hex
}

lhex_prereq() {
    info_prereq 'lhex'
    [ -x ../cmeta/hex ] \
 || hex
}

lhex() {
    lhex_prereq
    info_build 'lhex'
    cd ../cmeta
    ./hex < lhex.hex > lhex
    chmod +x lhex
}

elfwrap_prereq() {
    info_prereq 'elfwrap'
    [ -x ../cmeta/lhex ] \
 || lhex
}

elfwrap() {
    elfwrap_prereq
    info_build 'elfwrap'
    cd ../cmeta
    ./lhex < elfwrap.lhex > elfwrap
    chmod +x elfwrap
}

cmeta_prereq() {
    info_prereq 'cmeta'
    [ -x ../cmeta/lhex ] \
 || lhex

    [ -x ../cmeta/elfwrap ] \
 || elfwrap
}

cmeta_clean() {
    info_clean 'cmeta'
    cd ../cmeta
    rm -f cmeta cma asmrt cmeta1 cmrt cma1 elfwrap lhex
}

cmeta() {
    cmeta_prereq
    info_build 'cmeta'
    cd ../cmeta
    cat cma1.lhex rwx.lhex | ./lhex | ./elfwrap > cma1
    chmod +x cma1
    cat cmrt.lhex checkerror.lhex getlabel.lhex io.lhex rwx.lhex why.lhex label-z.lhex | ./lhex > cmrt
    ./cma1 < cmeta1.cma | cat cmrt - | ./elfwrap > cmeta1
    chmod +x cmeta1
    cat asmrt.lhex checkerror.lhex getlabel.lhex io.lhex rwx.lhex why.lhex label-z.lhex | ./lhex > asmrt
    ./cmeta1 < cma.cmeta | ./cma1 | cat asmrt - | ./elfwrap > cma
    chmod +x cma
    ./cmeta1 < cmeta.cmeta | ./cma | cat cmrt - | ./elfwrap > cmeta
    chmod +x cmeta
}

# -------------------- mcss --------------------

mcss_prereq() {
    info_prereq 'mcss'
    [ -x ../cmeta/cmeta ] \
 && [ -x ../cmeta/cma ] \
 && [ -e ../cmeta/cmrt ] \
 && [ -e ../cmeta/asmrt ] \
 || cmeta

    [ -x ../cmeta/elfwrap ] \
 || elfwrap
}

mcss_clean() {
    info_clean 'mcss'
    cd ../mcss
    rm -f ama2 am1 ama1 am1bm2 bm2 bm2bm1 bm2rt bma2 bm1 bm1cm3 \
        cma3 asmrt4 cm3 cm3rt cma2 cm2
}

mcss() {
    mcss_prereq
    info_build 'mcss'
    cd ../mcss
    ../cmeta/cmeta < cm2.cmeta | ../cmeta/cma | cat ../cmeta/cmrt - | ../cmeta/elfwrap > cm2
    chmod +x cm2
    ../cmeta/cmeta < cma2.cmeta | ../cmeta/cma | cat ../cmeta/asmrt - | ../cmeta/elfwrap > cma2
    chmod +x cma2
    cat cm3rt.lhex ../cmeta/io.lhex ../cmeta/rwx.lhex ../cmeta/why.lhex error.lhex getlabel.lhex put.lhex ../cmeta/label-z.lhex | ../cmeta/lhex > cm3rt
    ./cm2 < cm3.cm2 | ./cma2 | cat cm3rt - | ../cmeta/elfwrap > cm3
    chmod +x cm3
    cat asmrt4.lhex ../cmeta/io.lhex ../cmeta/rwx.lhex encode.lhex error.lhex getlabel.lhex put.lhex table.lhex ../cmeta/label-z.lhex | ../cmeta/lhex > asmrt4
    ./cm2 < cma3.cm2 | ./cma2 | cat asmrt4 - | ../cmeta/elfwrap > cma3
    chmod +x cma3
    ./cm3 < bm1.cm3 | ./cma3 | cat cm3rt - | ../cmeta/elfwrap > bm1cm3
    chmod +x bm1cm3
    ./bm1cm3 < bm1.bm1 | ./cma3 | cat cm3rt - | ../cmeta/elfwrap > bm1
    chmod +x bm1
    ./bm1 < bma2.bm1 | ./cma3 | cat asmrt4 - | ../cmeta/elfwrap > bma2
    chmod +x bma2
    cat bm2rt.lhex ../cmeta/io.lhex ../cmeta/rwx.lhex ../cmeta/why.lhex encode.lhex error.lhex getlabel.lhex put.lhex ../cmeta/label-z.lhex | ../cmeta/lhex > bm2rt
    ./bm1 < bm2.bm1 | ./bma2 | cat bm2rt - | ../cmeta/elfwrap > bm2bm1
    chmod +x bm2bm1
    ./bm2bm1 < bm2.bm1 | ./bma2 | cat bm2rt - | ../cmeta/elfwrap > bm2
    chmod +x bm2
    ./bm2 < am1.bm2 | ./bma2 | cat bm2rt - | ../cmeta/elfwrap > am1bm2
    chmod +x am1bm2
    ./bm1 < ama1.bm1 | ./cma3 | cat asmrt4 - | ../cmeta/elfwrap > ama1
    chmod +x ama1
    ./am1bm2 < am1.am1 | ./ama1 | cat bm2rt - | ../cmeta/elfwrap > am1
    chmod +x am1
    ./am1 < ama2.am1 | ./ama1 | cat asmrt4 - | ../cmeta/elfwrap > ama2
    chmod +x ama2
}

# -------------------- prox --------------------

prox_prereq() {
    info_prereq 'prox'
    [ -x ../mcss/am1 ] \
 && [ -x ../mcss/ama2 ] \
 && [ -e ../mcss/bm2rt ] \
 && [ -e ../mcss/asm4rt ] \
 || mcss

    [ -x ../cmeta/elfwrap ] \
 || elfwrap
}

prox_clean() {
    info_clean 'prox'
    cd ../prox
    rm -f prox2-asm prox2 prox2s1-asm prox2s1 prox1s2-asm prox1s2 elfwrap prox1-asm prox1
}

prox() {
    prox_prereq
    info_build 'prox'
    cd ../prox
    ../mcss/am1 < prox1.am1 | ../mcss/ama2 | cat ../mcss/bm2rt - | ../cmeta/elfwrap > prox1
    chmod +x prox1
    ../mcss/am1 < prox1-asm.am1 | ../mcss/ama2 | cat ../mcss/asmrt4 - | ../cmeta/elfwrap > prox1-asm
    chmod +x prox1-asm
    cat elfwrap.px1 rwx.px1 str.px1 | ./prox1 | ./prox1-asm | ../cmeta/elfwrap > elfwrap
    chmod +x elfwrap
    cat prox1s2.px1 compile.px1 num.px1 rwx.px1 str.px1 | ./prox1 | ./prox1-asm | ./elfwrap > prox1s2
    chmod +x prox1s2
    cat prox1s2-asm.px1 compile.px1 num.px1 rwx.px1 str.px1 table.px1 | ./prox1 | ./prox1-asm | ./elfwrap > prox1s2-asm
    chmod +x prox1s2-asm
    cat prox2s1.px1 compile.px1 num.px1 rwx.px1 str.px1 table.px1 | ./prox1s2 | ./prox1s2-asm | ./elfwrap > prox2s1
    chmod +x prox2s1
    cat prox2s1-asm.px1 compile.px1 num.px1 rwx.px1 str.px1 table.px1 | ./prox1s2 | ./prox1s2-asm | ./elfwrap > prox2s1-asm
    chmod +x prox2s1-asm
    cat prox2.px2 compile.px2 comprg.px2 num.px2 rwx.px2 str.px2 table.px2 | ./prox2s1 | ./prox2s1-asm | ./elfwrap > prox2
    chmod +x prox2
    cat prox2-asm.px2 compile.px2 comprg.px2 num.px2 rwx.px2 str.px2 table.px2 | ./prox2s1 | ./prox2s1-asm | ./elfwrap > prox2-asm
    chmod +x prox2-asm
}

# ------------------ asterix -------------------

asterix_prereq() {
    info_prereq 'asterix'
    [ -x ../prox/prox2 ] \
 && [ -x ../prox/prox2-asm ] \
 && [ -x ../prox/elfwrap ] \
 || prox
}

asterix_clean() {
    info_clean 'asterix'
    cd ../asterix
    rm -f asterix-asm asterix elfwrap asterix-s1-asm asterix-s1
}

asterix() {
    asterix_prereq
    info_build 'asterix'
    cd ../asterix
    cat asterix-s1.px2 ../prox/compile.px2 ../prox/num.px2 ../prox/rwx.px2 ../prox/str.px2 ../prox/table.px2 | ../prox/prox2 | ../prox/prox2-asm | ../prox/elfwrap > asterix-s1
    chmod +x asterix-s1
    cat asterix-s1-asm.px2 ../prox/compile.px2 ../prox/num.px2 ../prox/rwx.px2 ../prox/str.px2 ../prox/table.px2 | ../prox/prox2 | ../prox/prox2-asm | ../prox/elfwrap > asterix-s1-asm
    chmod +x asterix-s1-asm
    cat elfwrap.ax compile.ax error.ax rwx.ax | ./asterix-s1 | ./asterix-s1-asm | ../prox/elfwrap > elfwrap
    chmod +x elfwrap
    cat asterix.ax compile.ax error.ax num.ax rwx.ax table.ax | ./asterix-s1 | ./asterix-s1-asm | ./elfwrap > asterix
    chmod +x asterix
    cat asterix-asm.ax compile.ax error.ax num.ax rwx.ax table.ax | ./asterix-s1 | ./asterix-s1-asm | ./elfwrap > asterix-asm
    chmod +x asterix-asm
}

# ---------- Build control functions -----------

checkenv() {
    info2 'Checking environment'
    require ../asterix/asterix.ax
    require ../prox/prox2.px2
    require ../mcss/am1.am1
    require ../cmeta/cmeta.cmeta
}

build() {
    [ -n "$1" ] || die "build function needs a target as argument"
    [ -x ./make.sh ] || cd "`dirname $0`"
    checkenv
    case "$1" in
        asterix) asterix ;;
        prox) prox ;;
        mcss) mcss ;;
        cmeta) cmeta ;;
        prereq) asterix_prereq ;;
        clean) asterix_clean ;;
        clean-prereq)
            prox_clean
            mcss_clean
            cmeta_clean
            ;;
        clean-all)
            asterix_clean
            prox_clean
            mcss_clean
            cmeta_clean
            ;;
        *) die "Invalid build target: \"$1\"" ;;
    esac
    built=1
}

# ----------- Main code for make.sh ------------

loglevel=2
built=0

while [ -n "$1" ]; do
    case "$1" in
        help|-h|--h|--he|--hel|--help)
            usage
            ;;
        -q|--q|--qu|--qui|--quie|--quiet)
            [ $loglevel -eq 1 ] && loglevel=0 || loglevel=1
            ;;
        -qq*)
            loglevel=0
            ;;
        *)
            build "$1"
            ;;
    esac
    shift
done

if [ $built -eq 0 ]; then
	build asterix
fi
